/*

PORT B0 X phase 1
PORT B1 X phase 2
PORT B2 X phase 3
PORT B3 X phase 4
PORT B4 X Limit
PORT B5 Pen up/down
PORT B6 xx (xtal)
PORT B7 xx (xtal)

PORT C0 Y phase 1
PORT C1 Y phase 2
PORT C2 Y phase 3
PORT C3 Y phase 4
PORT C4 Y Limit
PORT C5
PORT C6 XX (reset)
PORT C7 XX (DNE)

PORT D0 xx (RxD)
PORT D1 xx (TxD)
PORT D2 CTS
PORT D3
PORT D4 x (reserve)
PORT D5 x (reserve)
PORT D6 x (reserve)
PORT D7 x (reserve)


******************************************************************/

#include "binary.h"
/*
#define setCTS()    PORTD|=  b00000100
#define clearCTS()  PORTD|=(~b00000100)
*/
#include <avr/io.h>
#include "bresnham.h"
#include "usart.h"
//#include <ctype.h>
#include "string.h"

/*
  ctype.h is screwed so we have to make our own wheels
*/
  
#define isdigit(A) (((A)>='0')&&((A)<='9'))?1:0
#define toupper(A) (((A)>='a')&&((A)<='z'))?(A)-0x20:(A)


#define OUTPUT  1
#define INPUT   0

#define forward  1
#define backward 0

#define SensorX()  ((PINB&16)>0)?1:0
#define SensorY()  ((PINC&16)>0)?1:0

#define PenUp()    PORTB|=b00100000
#define PenDown()  PORTB&=(~b00100000); Delay(9000)

#define Relax()  PORTB&=b11110000; PORTC&=b11110000


typedef enum e_mainstate {
     Cmd,
     Pos_X,
     Pos_Y
} mainstate;

  unsigned char stateTableBackward[] = {  
      (3)+(b0001<<4),                    
      (0)+(b0010<<4),
      (1)+(b0100<<4),
      (2)+(b1000<<4)
   };  
   
  unsigned char stateTableForward[] = {  
      (1)+(b0001<<4),                    
      (2)+(b0010<<4),
      (3)+(b0100<<4),
      (0)+(b1000<<4)
   }; 

struct stepper4ph_t {      
      unsigned Port       :3;
      unsigned MotorNum   :1;
      unsigned State      :3;
      unsigned Direction  :1;
} ;

typedef struct stepper4ph_t stepper4ph;

// -------- function declerations ------------
   
void goStep(int x, int y);
void gotoZero( void );
void runmotor (stepper4ph * motor, int speed, int distance);
void stepmotor (stepper4ph * motor);
void Delay(int delay);
void inline changeIO(char port, char mask, char bits);
void lineTo(DevPt dest_pt);

char CommandChar(char c) ;
char x_cordChar(char c);
char y_cordChar(char c);

void lineToXY(int X, int Y);


// -------| Global variables |-------

  mainstate   state;
  char        buff[3];
  stepper4ph  motor1, motor2; // "make a motor"
  DevPt       src_pt;
  int         x, y;
  

// -----| program |------

int main () {

   char c;

 
    // set up directions 
//    DDRA = (INPUT << PA0 | INPUT << PA1 |INPUT << PA2 |INPUT << PA3 |INPUT << PA4 |INPUT << PA5 |INPUT << PA6 |INPUT << PA7);
    DDRB = (OUTPUT << PB0 | OUTPUT << PB1 |OUTPUT << PB2 |OUTPUT << PB3 |INPUT << PB4 |OUTPUT << PB5 |INPUT << PB6 |INPUT << PB7);
    DDRC = (OUTPUT << PC0 | OUTPUT << PC1 |OUTPUT << PC2 |OUTPUT << PC3 |INPUT << PC4 |OUTPUT << PC5 |INPUT << PC6 /*|INPUT << PC7*/);
    DDRD = (INPUT << PD0 | INPUT << PD1 |INPUT << PD2 |INPUT << PD3 |INPUT << PD4 |INPUT << PD5 |INPUT << PD6 |INPUT << PD7);

  motor1.Port     = 1; // 0 = portA, so this is port B
  motor1.MotorNum = 0; // 0 is bits 0-3, 1 is bits 4-7

  motor2.Port     = 2;  // port C
  motor2.MotorNum = 0;  // bits 0-3

  USART_Init( 25 ); // 9600 @ 4Mhz
 
  setCTS();
 
  PenUp( );     
  gotoZero( );
  state = Cmd;
 
 c = USART_Receive();
 
  while(1){
    switch (state) {
      case Cmd:
        c = CommandChar(c);
      break;
      
      case Pos_X:
        c = x_cordChar(c);
      break;
      
      case Pos_Y:
        c = y_cordChar(c);
      break;
    
    }
    
  }
 
  Relax();
  
}

#define digitValue(A) ((A)-'0')

char x_cordChar(char c) {

  if (c == ',') {
    y = 0;
    state = Pos_Y;
    c = USART_Receive(); // time for another character
  } else if (!isdigit(c)) {
    state = Cmd;
  } else {  

    USART_Transmit(c);

    x *= 10;
    x += digitValue(c);
    c = USART_Receive(); // time for another character
  }
  
  return c;  
  
}

char y_cordChar(char c) {

  if (c == ',') { 
    lineToXY(x, y);
    x = 0;   
    state = Pos_X;
    c = USART_Receive(); // time for another character
  } if (!isdigit(c)) {
    state = Cmd;
    lineToXY(x, y);
  } else {
  
   USART_Transmit(c);
  
    y *= 10;
    y += digitValue(c);
    c = USART_Receive(); // time for another character
  }

  return c;
  
}

#define clearbuff()       buff[0] = '\0'; buff[1] = '\0'

char CommandChar(char c) {

  
  if (isdigit(c)) {
    x = 0;
    state = Pos_X;
  } else {    

    c = toupper(c);  

    if (strchr("SPUD", c)) {  // characters to accept

      buff[0] = buff[1];      
      buff[1] = c;    
      buff[2] = '\0';

      USART_Transmit(c);

      if (0) {    
      } else if (strcmp("PU", buff) == 0) {
	clearbuff();
	PenUp();    
      } else if (strcmp("PD", buff) == 0) {
	clearbuff();
	PenDown();    
      } else if (strcmp("SP", buff) == 0) {  //ignore the sp commands
	clearbuff();    
      }      

    } 
    c = USART_Receive(); // time for another character
  }

  return c;

}



/*
 this will go 1 of 8 directons, max 1 step in any of 2 or less directions
*/
void goStep(int x, int y) {

  if (x != 0) {
    if (x < 0) {
      motor1.Direction = forward;
    } else {
      motor1.Direction = backward;
    }
    stepmotor(&motor1);
  }

  if (y != 0) {
    if (y < 0) {
      motor2.Direction = forward;
    } else {
      motor2.Direction = backward;
    }
    stepmotor(&motor2);
  }

}

void lineToXY(int X, int Y) {
  DevPt temp_pt;
  double xc, yc;

  xc = (double) X / 1.69 ;
  yc = (double) Y / 3.4; 

  temp_pt.x = yc;
  temp_pt.y = xc;

  lineTo( temp_pt );  

}

void lineTo(DevPt dest_pt) {
  DevPt *pp;

  pp = bresenham_init(&src_pt, &dest_pt);
  do {	
    goStep((src_pt.x - pp->x), (src_pt.y - pp->y));
    src_pt.x = pp->x;
    src_pt.y = pp->y;  
    Delay(600);		
  } while (bresenham_next() != BRESENHAM_ERR);
  src_pt.x = dest_pt.x;
  src_pt.y = dest_pt.y;
  	
}


void gotoZero( void ) {

   motor1.Direction = backward;
   motor2.Direction = backward;   
   
   while(!SensorX() || !SensorY()) {    
     if(!SensorX())   stepmotor(&motor1);
     if(!SensorY())   stepmotor(&motor2);     
     Delay(600);      
   }
   
   motor1.Direction = forward;
   motor2.Direction = forward;
   while(SensorX() || SensorY()) {    
     if(SensorX())   stepmotor(&motor1);
     if(SensorY())   stepmotor(&motor2);     
     Delay(1000);      
   }
   
   src_pt.x = 0; src_pt.y = 0;

}

void runmotor (stepper4ph * motor, int speed, int distance) {

  for( ; distance != 0; distance --) {
    stepmotor(motor);
    Delay(speed);  
  }  
  
}

void stepmotor (stepper4ph * motor) {
          
   unsigned const char *  stateTable;
   unsigned char temp, mask;
   
   stateTable = stateTableBackward;                      //choose state table
   if (motor->Direction == forward) {
      stateTable = stateTableForward;
   } 
   
   motor->State  = (*(stateTable + motor->State) & 0x0F); // execute state machine
   temp =  (*(stateTable + motor->State) & 0xF0);
   
   mask = 0x0F;                                           // choose bit positions
   if (motor->MotorNum == 0) { temp >>= 4; mask = 0xF0; }
  
   changeIO( motor->Port, mask, temp);                    // change io bits
  
  }

void inline changeIO(char port, char mask, char bits) {
 switch (port) {
     case 0:
  //      PORTA = (PORTA&mask)|bits;
        break;
     case 1:
        PORTB = (PORTB&mask)|bits;
        break;
     case 2:
        PORTC = (PORTC&mask)|bits;
        break;
     case 3:
        PORTD = (PORTD&mask)|bits;
        break;
    }
}

void Delay(int delay) {
  int x;
  for (x = delay; x != 0; x--) {
    asm volatile ("nop"::); 
  }
}

